# meds/kernel.py
#
#

""" runtime namespace. """

from meds.utils.trace import get_exception
from meds.utils.misc import exclude, include
from meds.utils.type import iscmnd
from meds.errors import ESET, ENOTSET, ERESERVED
from meds.scheduler import Scheduler
from meds.launcher import Launcher
from meds.storage import Storage
from meds.object import Object
from meds.ool import OOL
from meds.cfg import cfg

import importlib
import logging
import pkgutil
import types
import sys
import os

class Kernel(Storage, Scheduler, Launcher):

    """ bot's kernel. """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._nr = 0
        self.names = OOL()
        self.cmnds = OOL()
        self.fleet = OOL()
        self.objs = Object()
        self.table = Object()

    def boot(self):
        """ initialize the kernel. """
        self.walk()
        self.ready()

    def announce(self, txt):
        for bots in self.fleet.values():
            for bot in bots: bot.announce(txt)

    def dispatch(self, event):
        """ launch a thread to handle an event. """
        self._nr += 1
        if not event.parsed: event.parse()
        if event.cmnd in ["init", "shutdown"]: logging.warn("ERESERVED %s" % event.cmnd) ; return []
        logging.info("# dispatch %s %s" % (event.origin, event.cmnd))
        if not self.names: self.walk()
        names = self.names.get(event.cmnd, [])
        for name in names:
            if not self.table.get(name, None): self.table[name] = self.direct(name)
        cmnds = self.cmnds.get(event.cmnd, [])
        thrs = []
        for cmnd in cmnds:
            thr = self.launch(cmnd, event)
            thrs.append(thr)
        return thrs

    def find(self, cmnd):
        """ command lookup. """
        names = self.names.get(cmnd, None)
        for name in names:
            mod = self.table[name]
            obj = getattr(mod, cmnd, None)
            if obj and type(obj) == types.FunctionType:
                if "event" in obj.__code__.co_varnames: yield obj

    def direct(self, name, package=None):
        """ directly import a module. """
        logging.info("! load %s" % name)
        return importlib.import_module(name, package)

    def reload(self, name, force=False):
        """ reload/load a module. """
        from meds.event import Event
        e = Event()
        n = name.split(".")[-1]
        if name in self.table and "shutdown" in dir(self.table[name]): self.table[name].shutdown(e) ; e.ready()
        if name not in self.table or force: self.table[name] = self.direct(name)
        else: self.table[name] = importlib.reload(self.table[name])
        if "init" in dir(self.table[name]): self.table[name].init(e) ; e.ready()
        return self.table[name]

    def modules(self):
        """ look for all modules that can be loaded. """
        for name in cfg.main.packages:
            package = self.direct(name)
            for pkg in pkgutil.walk_packages(package.__path__, name + "."): yield pkg[1]

    def walk(self):
        """ load all modules. """
        from meds.event import Event
        event = Event()
        thrs = []
        for name in sorted(list(self.modules()), reverse=True):
            mod = self.direct(name)
            for key in dir(mod):
                obj = getattr(mod, key, None)
                if obj and type(obj) == types.FunctionType:
                    if "event" in obj.__code__.co_varnames:
                        self.names.register(key, name)
                        self.cmnds.register(key, obj)
                if key == "init" and not exclude(name): thr = self.launch(mod.init, event) ; thrs.append(thr)
        self.waiter(thrs)
        self.ready()

    def list(self, want=""):
        """ list all functions in modules loaded with the bot. """
        for name in self.modules():
            if not include(name): continue
            if want and want not in name: continue
            mod = self.direct(name)
            for key in dir(mod):
                if key in ["init", "shutdown"]: continue
                obj = getattr(mod, key, None)
                if obj and type(obj) == types.FunctionType:
                    if "event" in obj.__code__.co_varnames: yield key

kernel = Kernel()
